<?php
if ( ! defined( 'ABSPATH' ) ) {
	exit; // Exit if accessed directly
}

use Braintree\Exception;
/**
 * Manager class that controls access to data stored in the database. 
 * @author Clayton Rogers
 * @since 3/12/16
 */
class Braintree_Manager{
	
	public static $_instance = null;
	public $required_settings;
	public $debug;
	public $log;
	public $braintreeSubscriptions;
	public $paypalOnly;
	public $version = '2.3.0';
	const Settings = 'braintree_payment_settings';
	
	/**
	 * Creates and instance of the Braintree_Manager class and loads all necessary data. 
	 */
	public function __construct(){
		$this->log = new Braintree_DebugLog();
		$this->init_settings();
		$this->required_settings = include_once WC_BRAINTREE_PLUGIN.'admin/includes/braintree-settings.php';
		$this->debug = $this->get_option('enable_debug') === 'yes' ? true : false;
		$this->initializeVariables();
		$this->initializeBraintree();
	}
	
	public static function instance(){
		if(self::$_instance === null){
			Braintree_Manager::$_instance = new Braintree_Manager();
		}
		return self::$_instance;
	}
	
	/**
	 * Set values for any class variables that are needed. 
	 */
	private function initializeVariables(){
		$this->braintreeSubscriptions = $this->get_option('braintree_subscriptions') === 'yes' ? true : false;
		$this->paypalOnly = $this->get_option('paypal_only') === 'yes' ? true : false;
		$this->cancelOnFail = $this->get_option('braintree_subscriptions_cancel_on_fail') === 'yes' ? true : false;
	}
	
	/**
	 * Load the settings values using the wordpress function get_options. 
	 */
	private function init_settings(){
		$this->settings = $this->get_payments_config(self::Settings);
	}
	
	public function isActive($option){
		return $this->get_option($option) === 'yes' ? true : false;
	}
	
	public function get_option($key){
		if($this->settings == null){
			$this->init_settings();
		}
		if(! isset($this->settings[$key])){
			$this->settings[$key] = isset($this->required_settings[$key]['default'])
			? $this->required_settings[$key]['default'] : '';
		}
		return $this->settings[$key];
	}
	
	public function set_option($key, $value = ''){
		$this->settings[$key] = $value;
	}
	
	public function update_settings($settings = null){
		if($this->settings == null){
			$this->init_settings();
		}
		if($settings != null){
			$this->settings = $settings;
		}
		$this->update_payments_config(self::Settings, $this->settings);
	}
	
	private function update_payments_config($option, $value){
		update_option(base64_encode($option), base64_encode(maybe_serialize($value)));
	}
	
	public function get_payments_config($option){
		return maybe_unserialize(base64_decode(get_option(base64_encode($option))));
	}
	
	public function delete_payments_config($option){
		delete_option(base64_encode($option));
	}
	
	public function display_debugLog(){
		$log = new Braintree_DebugLog();
		return $log->display_debugLog();
	}
	
	public function activateLicense($license_key){
		$args = array('timeout'=>60);
		$url_args = array('slm_action'=>'slm_activate',
			'secret_key'=>BRAINTREE_LICENSE_VERIFICATION_KEY,
			'license_key'=>$license_key);
		$response = wp_remote_get(add_query_arg($url_args, BRAINTREE_LICENSE_ACTIVATION_URL), $args);
		if($response instanceof WP_Error){
			foreach($response->get_error_messages() as $error)
			$this->log->writeErrorToLog('Error activating license: Message - '.$error);
		}
		else{
			if(isset($response['body'])){
				$response = json_decode($response['body'], true);
				if($response['result'] === 'success'){
					$this->set_option('license_status', 'active');
					$this->set_option('license', $license_key);
					$this->update_settings();
					$this->log->writeToLog('License activated. Message - '.$response['message']);
					add_action('admin_notices', __CLASS__.'::licenseActivationSuccessNotice');
				}
				else{
					$this->log->writeErrorToLog('There was an error activating your license. Message - '.$response['message']);
					add_action('admin_notices', __CLASS__.'::licenseActivationFailureNotice');
				}
			}
		}
	}
	
	public static function licenseActivationSuccessNotice(){
			echo '<div class="notice notice-success">
			  <p>'.__('Your license has been activated!', 'braintree').'</p>
			</div>';
	}
		
	public static function licenseActivationFailureNotice(){
			echo '<div class="notice notice-error">
			  <p>'.__('Your license could not be activated. Please check the debug log to view the error 
			  		message.', 'braintree').'</p></div>';
	}
	
	/**
	 * Deletes the debug log entries.
	 */
	public function deleteDebugLog(){
		$log = new Braintree_DebugLog();
		$log->delete_log();
	}
	
	/**
	 * Initialize all of the Braintree settings.
	 */
	private function initializeBraintree(){
		Braintree_Configuration::environment($this->getEnvironment());
		Braintree_Configuration::publicKey($this->get_option($this->getEnvironment().'_public_key'));
		Braintree_Configuration::privateKey($this->get_option($this->getEnvironment().'_private_key'));
		Braintree_Configuration::merchantId($this->get_option($this->getEnvironment().'_merchant_id'));
	}
	
	/**
	 * Generate a client token using the Braintree SDK. If a user has a customerId saved in the database, 
	 * then the client token will be generated using the vault customerId.
	 * @param array $args
	 */
	public function getClientToken($user_id = false){
		$clientToken = null;
		try{
			if($user_id){
				if($customerId = $this->getBraintreeCustomerId($user_id)){
					$clientToken = Braintree_ClientToken::generate(array('customerId'=>$customerId));
				}
				else{
					$clientToken = $this->getClientToken();
				}
			}
			else{
				$clientToken = Braintree_ClientToken::generate();
			}
		}catch(InvalidArgumentException $e){
			$this->log->writeErrorToLog($e->getMessage());
			if($customerId){
				delete_user_meta($user_id, 'braintree_vault_id');
				$clientToken = $this->getClientToken();
			}
		}
		catch(Exception $e){
			$this->log->writeErrorToLog('There was an error generating the client token. Message - '.$e->getMessage());
			return $clientToken;
		}
		return $clientToken;
	}
	
	/**
	 * Perform the Braintree_Transaction::sale() for the given $attribs and $order. If the sale is successfull, the transaction data is 
	 * saved to the order_meta data in the database. 
	 * @param array $args
	 * @param int $order_id
	 */
	public function woocommerceSale($attribs = array(), WC_Order $order){
		try{
			$response = Braintree_Transaction::sale($attribs);
			if($response->success){
				//update transaction meta in database.
				$this->saveTransactionMeta($response->transaction, $order);
				if($this->subscriptionsActive() && wcs_order_contains_subscription($order)){
					$subscriptions = wcs_get_subscriptions_for_order($order);
					foreach($subscriptions as $subscription){
						$this->saveTransactionMeta($response->transaction, $subscription);
						$this->updateOrderStatus($subscription);
					}
				}
				$this->updateOrderStatus($order);
				$this->log->writeToLog(sprintf('Transaction %s successfully charged in the amount of %s for orderId %s',
						$response->transaction->id, $order->get_total(), $order->id));
				$result = $this->handleResponseSuccess($order);
			}
			else{
				$result = $this->handleResponseError($response);
			}
			
		}catch(Braintree_Exception $e){
			if($e instanceof Braintree_Exception_Authorization){
				$this->log->writeErrorToLog('There was an Authorization Exception thrown during the transaction. Please check your 
						API key settings. Also, if you have configured a Merchant Account ID for settlement currency, verify you have entered the correct
						Merchant Account ID.');
				$e = new Exception(__('There was an error processing your payment.', 'braintree'));
			}
			else {
				$this->log->writeErrorToLog($e->getMessage());
			}
			$this->handleResponseError($e);
			return array('result'=>'failure', 'redirect'=>'');
		}
		return $result;
	}
	
	/**
	 * Process the Braintree_Transaction::sale().
	 * @param array $attribs
	 */
	public function sale(array $attribs){
		$response = null;
		try{
			$response = Braintree_Transaction::sale($attribs);
			if(! $response->success){
				$this->log->writeToLog($response->message);
			}
		}catch(Exception $e){
			$this->log->writeErrorToLog($e->getMessage);
			$response = $e;
		}
		return $response;
	}
	
	/**
	 * Update the order/subscriptions status.
	 * @param WC_Order $order
	 */
	public function updateOrderStatus(WC_Order $order){
		if($this->orderIsSubscription($order)){
			$order->update_status($this->get_option('subscriptions_payment_success_status'));
		}
		else{
			$order->update_status($this->get_option('order_status'));
		}
	}
	
	/**
	 * 
	 * @param Braintree $response
	 */
	public function handleResponseError($response){
		if($response instanceof Exception){
			if(function_exists('wc_add_notice')){
				wc_add_notice($response->getMessage(), 'error');
			}
			$this->log->writeErrorToLog($response->getMessage());
			return array('result'=>'failure', 'redirect'=>'');
		}
		else{
			if($response->errors->deepSize() > 0){
				foreach($response->errors->deepAll() as $error){
					if(function_exists('wc_add_notice')){
						wc_add_notice($error->message, 'error');
					}
					$this->log->writeErrorToLog($error->message);
				}
			}
			else{
				if(function_exists('wc_add_notice')){
					wc_add_notice($response->message, 'error');
				}
				$this->log->writeErrorToLog($response->message);
			}
			return array('result'=>'failure', 'redirect'=>'');
		}
	}
	
	public function handleSubscriptionResponseError($response){
		if($response instanceof Exception){
			if(current_user_can('manager_options')){
				wcs_add_admin_notice(sprintf('The subscription could not be cancelled. 
					Message: %s', $response->getMessage()), 'error');
			}
			$this->log->writeErrorToLog(sprintf('The subscription could not be cancelled. 
					Message: %s', $response->getMessage()));
		}
		else{
			if($response->errors->deepSize() > 0){
				foreach($response->errors->deepAll() as $error){
					if(current_user_can('manager_options')){
						wcs_add_admin_notice($error->message, 'error');	
					}
					$this->log->writeErrorToLog($error->message);
				}
			}
			else{
				if(current_user_can('manager_options')){
					wcs_add_admin_notice($response->message, 'error');	
				}
				$this->log->writeErrorToLog($response->message);
			}
			return array('result'=>'failure', 'redirect'=>'');
		}
	}
	
	public function handleSubscriptionCancellation($subscription){
		if(current_user_can('manage_options')){
			if(function_exists('wcs_add_admin_notice')){
				wcs_add_admin_notice(__('Subscription '.$subscription->id.' was successfully cancelled.', 'braintree'), 'success');
			}
		}
		$this->log->writeToLog(sprintf('Subscription %s was successfully cancelled.', $subscription->id));
	}
	
	/**
	 * Process the refund for the given order_id.
	 * @param int $order_id
	 * @param float $amount
	 * @param string $reason
	 */
	public function refund($order_id, $amount, $reason){
		$order = wc_get_order($order_id);
		try{
			$response = Braintree_Transaction::refund($order->get_transaction_id(), $amount);
			if($response->success){
				$this->log->writeToLog(sprintf('Order %s was refunded in the amount of %s', $order_id, $amount));
				return true;
			}
			else{
				$this->handleResponseError($response);
				throw new Exception(sprintf('A refund could not be processed for Order %s. Message: %s', $order_id, $response->message));
			}
		}catch(Exception\NotFound $e){
			$this->log->writeErrorToLog(sprintf('Refund could not be processed for Order %s because it was not found
					in the Braintree system. Message: %s.', $order_id, $e->getMessage()));
			throw new Exception(sprintf('Order %s could not be refunded because transaction ID %s could not be found in the Braintree Environment.', $order_id, $order->get_transaction_id()));
		}
	}
	
	/**
	 * Save the transaction data to the order meta in the database. 
	 * @param Braintree_Transaction $transaction
	 * @param WC_Order $order_id
	 */
	public function saveTransactionMeta($transaction, $order){
		update_post_meta($order->id, '_transaction_id', $transaction->id);
		update_post_meta($order->id, '_payment_method_title', $this->getPaymentMethodTitle($transaction));
		update_post_meta($order->id, '_payment_method_token', $this->getPaymentMethodToken($transaction));
		if($this->orderIsSubscription($order)){
			update_post_meta($order->id, '_subscription_type', 'woocommerce');
			update_post_meta($order->order->id, '_subscription_type', 'woocommerce');
			/*Save the transaction data to the parent order.*/
			$this->saveTransactionMeta($transaction, $order->order);
		}
	}
	
	/**
	 * Return the payment method title for the given transaction.
	 * @param Braintree_Transaction | Braintree_PaymentMethod $object
	 */
	public function getPaymentMethodTitle($object){
		$paymentTitle = '';
		if($object instanceof Braintree_Transaction){
			if($object->paymentInstrumentType === 'credit_card'){
				$paymentTitle = $object->creditCardDetails->cardType.' - '.$object->creditCardDetails->maskedNumber;
			}
			elseif($object->paymentInstrumentType === 'paypal_account'){
				$paymentTitle = 'PayPal - '.$object->paypalDetails->payerEmail;
			}
		}
		elseif($object instanceof Braintree_CreditCard){
			$paymentTitle = $object->cardType.' - '.$object->maskedNumber;
		}
		elseif($object instanceof  Braintree_PayPalAccount){
			$paymentTitle = 'PayPal - '.$object->email;
		}
		return $paymentTitle;
	}
	
	/**
	 * Return the token representing the payment method.
	 * @param Braintree_Transaction $transaction
	 */
	public function getPaymentMethodToken(Braintree_Transaction $transaction){
		if($transaction->paymentInstrumentType === 'credit_card'){
			$token = $transaction->creditCardDetails->token;
		}
		elseif($transaction->paymentInstrumentType === 'paypal_account'){
			$token = 'PayPal - '.$transaction->paypalDetails->token;
		}
		return $token;
	}
	
	/**
	 * Retrieve the parameter from the $_POST, $_GET, or $_REQUEST
	 * @param string $string
	 */
	public function getRequestParameter($string){
		$parameter;
		if(isset($_POST[$string])){
			$parameter = $_POST[$string];
		}
		elseif(isset($_GET[$string])){
			$parameter = $_GET[$string];
		}
		elseif(isset($_REQUEST[$string])){
			$parameter = $_REQUEST[$string];
		}
		else{
			$parameter = null;
		}
		return $parameter;
	}
	
	/**
	 * Create the Braintree Customer in the vault.
	 * @param number $userId
	 */
	public function createBraintreeCustomer($userId = 0, $attribs){
		try{
			$response = Braintree_Customer::create($attribs);
			if($response->success){
				update_user_meta($userId, 'braintree_vault_id', $response->customer->id);
				$this->log->writeToLog(sprintf('Braintree customerId %s created successfully for userId %s', $response->customer->id, $userId));
			}
			else{
				if($response->errors->deepSize() > 0){
					foreach($response->errors->deepAll() as $error){
						$this->log->writeErrorToLog($error->message);
					}
				}
				else{
					$this->log->writeErrorToLog($response->message);
				}
			}
		}catch(Exception $e){
			$this->log->writeErrorToLog(sprintf('There was an error while creating the braintree customer for user 
					%s. Message: %s', $userId, $e->getMessage()));
		}
	}
	
	public function getBraintreeCustomerId($userId){
		return get_user_meta($userId, 'braintree_vault_id', true);
	}
	
	/**
	 * Process the WooCommerce Subscription using the order_id provided. The order may contain multiple subscriptions and each
	 * one will be processed. 
	 * @param WC_Order $order_id
	 */
	public  function processWooCommerceSubscriptions(WC_Order $order){
		$result = null;
		$subscriptions = wcs_get_subscriptions_for_order($order);
		$result = $this->createWooCommerceSubscription($order);
		if($result['result'] === 'failure'){
			return $result;
		}
		else{
			WC()->cart->empty_cart();
			$this->updateOrderStatus($order);
		}
		return $result;
	}
	
	/**
	 * Process the Braintree subscription using the order_id provided. The order may contain several 
	 * subscriptions and each one will be processed.  
	 * @param WC_Order $order
	 */
	public function processBraintreeSubscriptions($order){
		$subscriptions = wcs_get_subscriptions_for_order($order);
		/*Loop through each subscription and create it in the Braintree Environment. A single subscription
		can have multiple items. It is better to create a subscription for each item.*/ 
		foreach($subscriptions as $subscription){
			$result = $this->createBraintreeSubscription($subscription);
			if($result['result'] === 'failure'){
				return $result;
			}
		}
		$this->updateOrderStatus($order);
		return $result;
	}
	
	/**
	 * Create the Braintree Subscription.
	 * @param WC_Product $product
	 * @param int $index
	 */
	public function createBraintreeSubscription(WC_Subscription $subscription){
		$planId = $this->getSubscriptionPlanId($subscription);
		$result = null;
		if(! $planId){
			$result = $this->handleWCError(sprintf('Product %s does not have a valid planId.', $product->get_title()));
		}
		else{
			try{
				$attribs = array(
						'id'=>$subscription->id,
						'planId'=>$planId,
						'paymentMethodNonce'=>$this->getRequestParameter('payment_method_nonce'),
						'firstBillingDate'=>$this->getSubscriptionBillingDate($subscription),
				);
				$attribs = $this->getMerchantAccountId($attribs);
				$attribs = $this->populateSubscriptionPrice($attribs, $subscription);
				$response = Braintree_Subscription::create($attribs);
				if($response->success){
					$this->saveBraintreeSubscriptionMeta($subscription, $response->subscription);
					$this->updateOrderStatus($subscription);
					$this->updateCart($subscription);
					$result = $this->handleResponseSuccess($subscription);
				}
				else{
					$result = $this->handleResponseError($response);
				}
			}catch(Exception $e){
				$this->log->writeErrorToLog(sprintf('There was an error processing the subscription. 
						Message: %s', $e->getMessage()));
				return $this->handleWCError(sprintf('There was an error processing the subscription. 
						Message: %s', $e->getMessage()));
			}
		}
		return $result;
	}

	/**
	 * Update the Braintree_Subscription
	 * @param array $attribs
	 */
	public function updateBraintreeSubscription($id, $attribs){
		$result = null;
		try{
			$response = Braintree_Subscription::update($id, $attribs);
			$result = $response->success;
		}catch(Braintree_Exception_NotFound $e){
			$this->log->writeErrorToLog(sprintf('Braintree subscription %s could not be found.', $id));
			$result = false;
		}
		catch(Exception $e){
			$this->log->writeErrorToLog($e->getMessage());
			$result = false;
		}
		return $result;
	}
	
	/**
	 * Create the WooCommerce subscription.
	 * @param WC_Subscription $subscription
	 */
	public function createWooCommerceSubscription(WC_Order $subscription){
		$attribs = WC_Braintree_Payments::createOrderAttributes($subscription);
		$attribs['paymentMethodNonce'] = $this->getRequestParameter('payment_method_nonce');
		return $this->woocommerceSale($attribs, $subscription);
	}
	
	/**
	 * Cancel the Braintree subscription.
	 * @param WC_Subscription $subscription
	 * @param bool $update_status default is false. 
	 */
	public function cancelBraintreeSubscription(WC_Subscription $subscription, $update_status = false){
		$id = get_post_meta($subscription->id, '_subscription_id', true);
		try{
			$response = Braintree_Subscription::cancel($id);
			if($response->success){
				if($update_status){
					$subscription->update_status('cancelled');
				}
				$this->handleSubscriptionCancellation($response->subscription);
			}
			else{
				$this->handleSubscriptionResponseError($response);
			}
		}catch(Exception $e){
			$this->handleSubscriptionResponseError($e);
		}
		return $response->success;
	}
	
	/**
	 * Cancel the WooCommmerce Subscription.
	 * @param WC_Subscription $subscription
	 */
	public function cancelWooCommerceSubscription(WC_Subscription $subscription){
		$subscription->update_status('cancelled');
		return true;
	}
	
	/**
	 * Fetch the planId for the given product.
	 * @param WC_Subscription $subscription
	 */
	public function getSubscriptionPlanId(WC_Subscription $subscription){
		$planId = null;
		$items = $subscription->get_items();
		foreach($items as $item){
			$product_id = $item['product_id'];
			if(WC_Subscriptions_Product::is_subscription($product_id)){
				$planId = get_post_meta($product_id, 'braintree_subscription_id', true);
			}
		}
		return $planId;
	}
	
	/**
	 * Save the transaction data for the subscription to the database. The transcation data is saved to the subscription
	 * and the parent order. If there is not transaction, then a call is made to Braintree to fetch the payment method details.
	 * @param WC_Subscription $subscription
	 * @param Braintree_Subscription $braintree_subscription
	 */
	public function saveBraintreeSubscriptionMeta($subscription, $braintree_subscription){
		$transaction = isset($braintree_subscription->transactions[0]) ? $braintree_subscription->transactions[0] : null;
		update_post_meta($subscription->id, '_subscription_id', $braintree_subscription->id);
		update_post_meta($subscription->id, 'braintree_plan_id', $braintree_subscription->planId);
		update_post_meta($subscription->id, '_payment_method_token', $braintree_subscription->paymentMethodToken);
		update_post_meta($subscription->id, '_subscription_type', 'braintree');
		update_post_meta($subscription->order->id, '_subscription_type', 'braintree');
		if($transaction){
			update_post_meta($subscription->id, '_transaction_id', $transaction->id);
			$this->saveTransactionMeta($transaction, $subscription->order);
			update_post_meta($subscription->id, '_payment_method_title', $this->getPaymentMethodTitle($transaction));
		}
		/*If there is not transaction, then fetch the payment method  using the token and use it to save the payment method title.*/
		else{
			$paymentMethod = $this->getBraintreePaymentMethod($braintree_subscription->paymentMethodToken);
			update_post_meta($subscription->id, '_payment_method_title', $this->getPaymentMethodTitle($paymentMethod));
			update_post_meta($subscription->order->id, '_payment_method_title', $this->getPaymentMethodTitle($paymentMethod));
		}
	}
	
	/**
	 * Save the transaction data for the subscription to the database.
	 * @param WC_Subscription $subscription
	 * @param Braintree_Subscription $braintree_subscription
	 */
	public function saveWooCommerceSubscriptionMeta($subscription, $transaction){
		update_post_meta($subscription->id, '_transaction_id', $transaction->id);
		update_post_meta($subscription->id, '_payment_method_token', $braintree_subscription->paymentMethodToken);
		update_post_meta($subscription->id, '_payment_method_title', $this->getPaymentMethodTitle($transaction));
	}
	
	/**
	 * 
	 * @param string $message
	 */
	public function handleWCError($message){
		wc_add_notice($message, 'error');
		return array('result'=>'failure', 'redirect'=>'');
	}
	
	/**
	 * Returns an array o consiting of a result and redirect index. This method is to be used only for 
	 * order and subscription creation.
	 * @param WC_Order $order
	 */
	public function handleResponseSuccess($order){
		return array('result'=>'success', 'redirect'=>$order->get_checkout_order_received_url());
	}
	
	public function getEnvironment(){
		$environment = $this->get_option ( 'production_environment' ) === 'yes' ? 'production' : 'sandbox';
		if (strtolower ( $this->get_option ( 'license_status' ) ) === 'inactive') {
			$environment = 'sandbox';
		}
		return $environment;
	}
	
	public function isEnabled(){
		return $this->get_option('enabled') === 'yes' ? 'yes' : 'no';
	}
	
	/**
	 * Method that gets all of the confiured plan Id's in the Braintree environment. 
	 */
	public function getBraintreePlans(){
		$plans = array();
		try{
			$plans = Braintree_Plan::all();
		}catch(Exception $e){
			$this->log->writeErrorToLog($e->getMessage());
		}
		return $plans;
	}
	
	/**
	 * Retrieve the payment method from Braintree using the specified token.
	 * @param string $token
	 * @return 
	 */
	public function getBraintreePaymentMethod($token){
		$paymentMethod = null;
		try{
			$paymentMethod = Braintree_PaymentMethod::find($token);
			if(! $paymentMethod){
				throw new Exception($response->message);
			}
		}catch(Exception $e){
			$this->log->writeErrorToLog('Method getBraintreePaymentMethod(). Failed to retrieve customer 
					payment method. Message: '.$e->getMessage);
		}
		return $paymentMethod;
	}
	
	/**
	 * Retrieve the Braintree Customer from Braintree. 
	 * @param string $customerId
	 * @return null | Braintree_Customer $customer
	 */
	public function getBraintreeCustomer($customerId){
		$customer = null;
		try{
			$customer = Braintree_Customer::find($customerId);
		}catch(Exception $e){
			$this->log->writeErrorToLog(sprintf('Braintree customer %s could not be found. 
					Message: %s', $customerId, $e->getMessage()));
			return $customer;
		}
		return $customer;
	}
	
	/**
	 * Create the payment method in Braintree using the provided nonoce. 
	 * @param string $nonce
	 */
	public function createPaymentMethod($nonce, $customerId){
		$paymentMethod = null;
		try{
			$response = Braintree_PaymentMethod::create(array(
					'customerId'=>$customerId,
					'paymentMethodNonce'=>$nonce
			));
			if($response->success){
				$paymentMethod = $response->paymentMethod;
			}
			else{
				throw new Exception($response->message);
			}
		}catch(Exception $e){
			$this->log->writeErrorToLog($e->getMessage());
			return $paymentMethod;
		}
		return $paymentMethod;
	}
	
	/**
	 * Return the first billing date for the subscription.
	 * @param WC_Subscription $subscription
	 */
	public function getSubscriptionBillingDate($subscription){
		if(! $timeStamp = $subscription->get_time('trial_end')){
			$timeStamp = $subscription->get_time('start');
		}
		$startDate = new DateTime();
		$startDate->setTimestamp($timeStamp);
		return $startDate->format('m/d/Y'); //ex. 01/01/2016
	}
	
	/**
	 * Get the formatted date mm/dd/yyyy for the specified date type
	 * @param WC_Subscription $subscription
	 * @param string $dateType
	 * @return string date
	 */
	public function getSubscriptionDate(WC_Subscription $subscription, $dateType){
		$timestamp = $subscription->get_time('next_payment');
		$date = new DateTime();
		$date->setTimestamp($date);
		return $date->format('m/d/Y');
	}
	
	/**
	 * Return the subscription price. 
	 * @param WC_Subscription $subscription
	 */
	public function getSubscriptionPrice($subscription){
		$items = $subscription->get_items();
		foreach($items as $item){
			$product_id = $item['product_id'];
			return get_post_meta($product_id, '_subscription_price', true);
		}
	}
	
	public function populateSubscriptionPrice($attribs, $subscription){
		$price = $this->getSubscriptionPrice($subscription);
		if(! empty($price)){
			$attribs['price'] = $price;
		}
		return $attribs;
	}
	
	/**
	 * Remove the subscription from the shopping cart. 
	 * @param WC_Subscription $subscription
	 */
	public function updateCart($subscription){
		foreach( WC()->cart->get_cart() as $cart_key => $item ){
			$product_id = $item['product_id'];
			foreach($subscription->get_items() as $item) {
				if($item['product_id'] == $product_id) {
					WC()->cart->remove_cart_item($cart_key);
				}
			}
		}
	}
	
	public function orderIsSubscription(WC_Order $order){
		$isSubscription = false;
		if($this->subscriptionsActive()){
			if($order instanceof WC_Subscription){
				$isSubscription = true;
			}
		}
		return $isSubscription;
	}
		
	public function subscriptionsActive(){
		return class_exists('WC_Subscription');
	}
	
	/**
	 * Retrieve the Merchant Account for the shop.
	 * @param array $attribs
	 * @return unknown
	 */
	public function getMerchantAccountId($attribs = array()){
		$currencyCode = get_woocommerce_currency();
		$merchantId = BT_Manager()->get_option('woocommerce_braintree_'.BT_Manager()->getEnvironment().'_merchant_account_id['.$currencyCode.']');
		if(! empty($merchantId)){
			$attribs['merchantAccountId'] = $merchantId;
		}
		return $attribs;
	}
	
	public function getBraintreeCustomerObject($attribs = array(), WC_Order $order){
			$attribs['customer'] = array(
						'email'=>$order->billing_email,
						'firstName'=>$order->billing_first_name,
						'lastName'=>$order->billing_last_name,
						'phone'=>$order->billing_phone
				);
		return $attribs;
	}
	
	public function getDropinContainer(){
		if($this->getEnvironment() === 'sandbox'){?>
		<div class="gateway-environment"><?php echo __('Sandbox Mode', 'braintree')?></div>
		<?php }?>
		<?php 
			$paymentMethods = $this->get_option('payment_methods');
			if(! empty($paymentMethods)){
				?>
				<div class="accepted-payment-methods">
				  <?php foreach($paymentMethods as $key=>$method){
				  		if(! empty(BT_Manager()->settings['payment_methods'][$key])){?>
				  			<div class="payment-method"><img src="<?php echo Braintree_PaymentMethods::paymentMethods()[$key]['src']?>"/></div>
				  <?php }
				  }?>
				</div>
				<?php
			}?>
		<div id="dropin-container"></div>	
	<?php 
	}
	
	public function getDonationDropinContainer(){
		if($this->getEnvironment() === 'sandbox'){?>
				<div class="gateway-environment"><?php echo __('Sandbox Mode', 'braintree')?></div>
				<?php 
					$paymentMethods = $this->get_option('donation_payment_methods');
					if(! empty($paymentMethods)){
						?>
						<div class="accepted-payment-methods">
						  <?php foreach($paymentMethods as $key=>$method){
						  		if(! empty(BT_Manager()->settings['donation_payment_methods'][$key])){?>
						  			<div class="payment-method"><img src="<?php echo Braintree_PaymentMethods::paymentMethods()[$key]['src']?>"/></div>
						  <?php }
						  }?>
						</div>
						<?php
					}
				}?>
				<div id="dropin-container"></div>	
			<?php 
	}
	
	public function getPayPalContainer(){
	?>
		<div id="dropin-container"></div>
	<?php 
	}
	
	public static function getSubscriptionStatuses(){
		$subscription_statuses = array(
				'wc-pending'        => _x( 'Pending', 'Subscription status', 'woocommerce-subscriptions' ),
				'wc-active'         => _x( 'Active', 'Subscription status', 'woocommerce-subscriptions' ),
				'wc-on-hold'        => _x( 'On hold', 'Subscription status', 'woocommerce-subscriptions' ),
				'wc-cancelled'      => _x( 'Cancelled', 'Subscription status', 'woocommerce-subscriptions' ),
				'wc-switched'       => _x( 'Switched', 'Subscription status', 'woocommerce-subscriptions' ),
				'wc-expired'        => _x( 'Expired', 'Subscription status', 'woocommerce-subscriptions' ),
				'wc-pending-cancel' => _x( 'Pending Cancellation', 'Subscription status', 'woocommerce-subscriptions' ),
		);
		return $subscription_statuses;
	}
}

/**
 * Function that returns an instance of the Braintree_Manager class. If there is no instance, then a new instance is
 * instantiated and asigned to the static variable $_instance of class Braintree_Manager.
 * @return BT_Manager
 */
function BT_Manager(){
	return Braintree_Manager::instance();
}