<?php
if ( ! defined( 'ABSPATH' ) ) {
	exit; // Exit if accessed directly
}
/**
 * 
 * @author Clayton Rogers
 * @copyright Payment Plugins
 * @since 3/14/2016
 */
class WC_Braintree_Payments extends WC_Payment_Gateway{
	
	const gatewayName = 'braintree_payment_gateway';
	
	public $supports = array('subscriptions','products','subscription_cancellation','multiple_subscriptions',
				'subscription_amount_changes','subscription_date_changes','default_credit_card_form','refunds',
				'pre-orders','subscription_payment_method_change_admin','gateway_scheduled_payments','subscription_payment_method_change_customer'
		);
	
	/**
	 * Initialize any data needed for payment processing with Braintree
	 */
	public function __construct(){
		$this->enabled = BT_Manager()->isEnabled();
		$this->id = self::gatewayName;
		$this->title = $this->get_option('title_text');
		$this->has_fields = true;
		add_action('wp_enqueue_scripts', __CLASS__.'::loadScripts');
	}
	
	public static function init(){
		add_action('woocommerce_before_checkout_form', __CLASS__.'::createBraintreeCustomer');
	}
	
	public static function loadScripts(){
		if(BT_Manager()->isEnabled() === 'yes'){
			wp_enqueue_script('braintree-dropin', BRAINTREE_DROPIN_JS, null, true);
			wp_enqueue_style('braintree-payments-style', WC_BRAINTREE_ASSETS.'css/braintree-payments.css');
		}
	}
	
	public static function addGateway($methods = array()){
		$methods[] = 'WC_Braintree_Payments';
		return $methods;
	}
	
	/**
	 * Process the payment for the order.
	 * {@inheritDoc}
	 * @see WC_Payment_Gateway::process_payment()
	 */
	public function process_payment($order_id){
		if(self::isPaymentChangeRequest()){
			$order = wc_get_order($order_id);
			return array('result'=>'success', 'redirect'=>$order->get_checkout_order_received_url());
		}
		if(self::subscriptionsActive() && $this->isSubscription($order_id)){
			return $this->processSubscription($order_id);
		}
		else{
			$nonce = self::getRequestParameter('payment_method_nonce');
			if(! $nonce){
				return self::wooCoomerceError(__('A valid payment nonce is required to process the payment.', 'braintree'));
			}
			$attribs = self::createOrderAttributes($order_id);
			$attribs['paymentMethodNonce'] = $nonce;
			$result = BT_Manager()->woocommerceSale($attribs, wc_get_order($order_id));
			if($result['result'] === 'success'){
				WC()->cart->empty_cart();
			}
			return $result;
		}
	}
	
	/**
	 * Retrieve the payment fields to be displayed on the checkout page. If the request is for a payment method change, 
	 * render the payment methods form. 
	 * {@inheritDoc}
	 * @see WC_Payment_Gateway::payment_fields()
	 */
	public function payment_fields(){
		if(self::isPaymentChangeRequest()){
			WC_Braintree_Subscription_Scripts::getPaymentChangeScript();
			echo WC_Braintree_Subscriptions::getPaymentMethodForm(wp_get_current_user()->ID);
			?>
			<div id="payment_form_container" class="payment-form-container">
			<?php if(get_user_meta(wp_get_current_user()->ID, 'braintree_payment_methods', true)){?>
			  <div class="payment-method-button"><span id="cancel_add_new"><?php echo __('Cancel', 'braintree')?></span></div>	
			<?php }?>
			  <?php BT_Manager()->getDropinContainer();?>
			</div>
			<script>
			var braintreeUtils = new BraintreeUtils("<?php echo BT_Manager()->getClientToken()?>");
			$(document).ready(function(){$(document.body).trigger('braintree_form_ready')});
			</script>
			<?php 
		}
		else if(self::isPayForOrder()){
			WC_Braintree_Checkout_Scripts::getPayForOrderScript();
			BT_Manager()->getDropinContainer();
			?>
			<script>
			var braintreeUtils = new BraintreeUtils("<?php echo BT_Manager()->getClientToken(wp_get_current_user()->ID)?>");
			$(document).ready(function(){$(document.body).trigger('braintree_form_ready')});
			</script>
			<?php 
		}
		else{
			if(BT_Manager()->paypalOnly){
				WC_Braintree_Checkout_Scripts::getPayPalScript();
				BT_Manager()->getPayPalContainer();
				?>
				<script>
  				var paypalCheckout = new PayPalCheckout("<?php echo BT_Manager()->getClientToken()?>");
  				$(document).ready(function(){$(document.body).trigger('paypal_container_ready')});
				</script>
				<?php 
			}
			else{
				WC_Braintree_Checkout_Scripts::getCheckoutScript();
				BT_Manager()->getDropinContainer();
				?>
				<script>
				var braintreeUtils = new BraintreeUtils("<?php echo BT_Manager()->getClientToken(wp_get_current_user()->ID)?>");
				$(document).ready(function(){$(document.body).trigger('braintree_form_ready')});
				</script>
	 <?php }?>	
 <?php }
	}
	
	/**
	 * Process the refund for the given $order_id.
	 * {@inheritDoc}
	 * @see WC_Payment_Gateway::process_refund()
	 */
	public function process_refund($order_id, $amount = null, $reason = ''){
		if(! $this->canRefundOrder($order_id)){
			BT_Manager()->log->writeToLog(sprintf('Order %s cannot be refunded because there is 
					not a transaction associated with the order. If this order is associated with a subscription
					that has not processed a payment yet, then there will be no transaction available until the subscription payment is processed. .', $order_id));
			return new WP_Error(404, __('The order cannot be refunded because there is no transaction ID associated with the order. Check the debug log for more details.', 'braintree'));	
		}
		try{
			BT_Manager()->refund($order_id, $amount, $reason);
		}catch(Exception $e){
			return new WP_Error(404, $e->getMessage());
		}
	}
	
	/**
	 * Create the Braintree Customer if none exists for the user.
	 */
	public static function createBraintreeCustomer(){
		global $current_user;
		get_currentuserinfo();
		if($userId = wp_get_current_user()->ID){
			if(! $customerId = get_user_meta($userId, 'braintree_vault_id')){
				$attribs = array();
				$attribs['firstName'] = get_user_meta($userId, 'first_name', true);
				$attribs['lastName'] = get_user_meta($userId, 'last_name', true);
				$attribs['email'] = $current_user->user_email;
				$attribs['phone'] = get_user_meta($userId, 'billing_phone', true);
				BT_Manager()->createBraintreeCustomer($userId, $attribs);
			}
		}
	}
	
	public function get_option($option, $default = null){
		return BT_Manager()->get_option($option);
	}
	
	/**
	 * Retrieve the parameter from the $_POST, $_GET, or $_REQUEST
	 * @param string $string
	 */
	public static function getRequestParameter($string){
		$parameter;
		if(isset($_POST[$string])){
			$parameter = $_POST[$string];
		}
		elseif(isset($_GET[$string])){
			$parameter = $_GET[$string];
		}
		elseif(isset($_REQUEST[$string])){
			$parameter = $_REQUEST[$string];
		}
		else{
			$parameter = null;
		}
		return $parameter;
	}
	
	/**
	 * Create an array of attributes for the WooComerce Order.
	 * @param WC_Order|int $order_id
	 */
	public static function createOrderAttributes($order){
		$attribs = array();
		if(! is_object($order)){
			$order = wc_get_order($order);
		}
		$attribs['amount'] = $order->get_total();
		$attribs['billing'] = array('countryCodeAlpha2'=>self::getRequestParameter('billing_country'),
				'firstName'=>self::getRequestParameter('billing_first_name'), 
				'lastName'=>self::getRequestParameter('billing_last_name'),
				'postalCode'=>self::getRequestParameter('billing_postcode'),
				'streetAddress'=>self::getRequestParameter('billing_address_1')
		);
		$attribs['options'] = array('submitForSettlement'=>true);
		$attribs['orderId'] = self::getOrderId($order);
		/*add Merchant Account if configured.*/
		$attribs = self::getMerchantAccountId($attribs);
		$attribs = self::getCustomerObject($attribs, $order);
		return $attribs;
		
	}
	
	/**
	 * Process the WC_Subscription. The subscription will be handled differently, based on whether or not 
	 * braintree_subscriptions or woocommerce_subscriptions has been enabled by the admin.
	 * @param int $order_id
	 */
	public function processSubscription($order_id){
		$result = null;
		$order = wc_get_order($order_id);
		if($this->get_option('braintree_subscriptions') === 'yes'){
			$result = BT_Manager()->processBraintreeSubscriptions($order);
		}
		else{
			$result = BT_Manager()->processWooCommerceSubscriptions($order);
		}
		return $result;
	}
	
	public static function getMerchantAccountId($attribs = array()){
		return BT_Manager()->getMerchantAccountId($attribs);
	}
	
	public static function getCustomerObject($attribs = array(), $order){
		return BT_Manager()->getBraintreeCustomerObject($attribs, $order);	
	}
	
	public static function wooCoomerceError($message){
		wc_add_notice($message, 'error');
		return array('result'=>'failure', 'redirect'=>'');
	}
	
	public static function isPaymentChangeRequest(){
		return (isset($_REQUEST['woocommerce_change_payment']) || 
			isset($_REQUEST['change_payment_method']));
	}
	
	public static function isPayForOrder(){
		return (isset($_REQUEST['pay_for_order']) && $_REQUEST['pay_for_order'] === 'true');
	}
	
	public static function subscriptionsActive(){
		return class_exists('WC_Subscription');
	}
	
	public function isSubscription($order_id){
		return WC_Braintree_Subscriptions::orderContainsSubscription($order_id);
	}
	
	/**
	 * Return the order prefix if one has been configured. 
	 * @param WC_Order $order
	 * @return string $orderId
	 */
	public static function getOrderId(WC_Order $order){
		$orderId = $order->id;
		if(self::subscriptionsActive() && ($order instanceof WC_Subscription)){
			$orderId = 	$orderId.BT_Manager()->get_option('woocommerce_subscriptions_prefix');
		}
		else{
			$orderId = $orderId.BT_Manager()->get_option('order_prefix');
		}
		return $orderId;
	}
	
	/**
	 * Method that checks if the order has a valid transaction Id.
	 * @param int $order_id
	 */
	public function canRefundOrder($order_id){
		return wc_get_order($order_id)->get_transaction_id();
	}
	
	public function admin_options(){
		?>
		<div class="admin-options-links">
		  <ul>
			<li><a href="<?php echo get_admin_url().'admin.php?page=braintree-payment-settings'?>"><?php echo __('API Settings', 'braintree')?></a></li>
			<li><a href="<?php echo get_admin_url().'admin.php?page=braintree-woocommerce-settings'?>"><?php echo __('WooCommerce Settings', 'braintree')?></a></li>
			<li><a href="<?php echo get_admin_url().'admin.php?page=braintree-subscriptions-page'?>"><?php echo __('Subscription Settings', 'braintree')?></a></li>
			<li><a href="<?php echo get_admin_url().'admin.php?page=braintree-donations-page'?>"><?php echo __('Donation Settings', 'braintree')?></a></li>
			<li><a href="<?php echo get_admin_url().'admin.php?page=braintree-webhooks-page'?>"><?php echo __('Webhook Settings', 'braintree')?></a></li>
			<li><a href="<?php echo get_admin_url().'admin.php?page=braintree-debug-log'?>"><?php echo __('Debug Log', 'braintree')?></a></li>
			<li><a href="<?php echo get_admin_url().'admin.php?page=braintree-license-page'?>"><?php echo __('License Activation', 'braintree')?></a></li>
		  </ul>
	   </div>
		<?php 
	}
	
}
add_filter('woocommerce_payment_gateways', 'WC_Braintree_Payments::addGateway');
WC_Braintree_Payments::init();